/*
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is Mixp.

The Initial Developer of the Original Code is Thierry Bzecourt.
Portions created by Thierry Bzecourt are Copyright (C) 1999, 2000
Thierry Bzecourt. All Rights Reserved.

Contributor(s):

Alternatively, the contents of this file may be used under the terms
of the GNU General Public License (the "GPL"), in which case the
provisions of the GPL are applicable instead of those above.  If you
wish to allow use of your version of this file only under the terms of
the GPL and not to allow others to use your version of this file under
the MPL, indicate your decision by deleting the provisions above and
replace them with the notice and other provisions required by the
GPL. If you do not delete the provisions above, a recipient may use
your version of this file under either the MPL or the GPL.
*/

/*

  Coding conventions:

  - expat functions and datatypes start with XML_, and contain one or
  more capitalized words.

  - gexpat C functions and datatypes start with xml_, and contain
  the same words as their expat counterparts, but separated by
  underscores and not capitalized.

  - gexpat Scheme functions and datatypes start with expat:,
  and contain the same words as their C counterparts, but separated by
  hyphens and not capitalized.

  "void *" expat datatypes have usually been mapped to smobs.

*/
 

#include "xmlparse.h"

#define GEXPAT_MODULE_NAME "xml expat"

void 
init_gexpat(void);

XML_Encoding *
xml_encoding_smob_to_expat(SCM xml_encoding_smob);

typedef struct {
  SCM data;

  short use_parser_as_handler_arg; /* 1 if the data if
				      xml_use_parser_as_handler_arg()
				      was called */
  SCM parser; /* defined only if the previous parameter is 1 */
  SCM external_entity_ref_handler_arg; /* only used when 
					  xml_set_external_entity_ref_arg()
					  is called */

  /* Maybe this structure is too big (could be implemented with a list
     which would contain only the implemented handlers ) */

  SCM start_element_handler;
  SCM end_element_handler;
  SCM character_data_handler;
  SCM processing_instruction_handler;
  SCM comment_handler;
  SCM start_cdata_section_handler;
  SCM end_cdata_section_handler;
  SCM default_handler;
  SCM default_handler_expand;
  SCM unparsed_entity_decl_handler;
  SCM notation_decl_handler;
  SCM start_namespace_decl_handler;
  SCM end_namespace_decl_handler;
  SCM not_standalone_handler;
  SCM external_entity_ref_handler;

} user_data;

/*
 * A structure used to wrap encoding handler data when calling
 * XML_SetUnknownEncodingHandler
 */ 
typedef struct {
  SCM data;     /* the encoding handler data provided by the Scheme caller */
  SCM handler;  /* a reference on the handler, used by
		   generic_unknown_encoding_handler */
} encoding_handler_data;
